"""
    BOARD utilities
"""
import cij.ssh
import cij

PREFIX = "BOARD"
REQUIRED = ["FORMFACTOR", "MEMORY", "CHIP", "ALIAS"]
EXPORTED = ["CLASS", "IDENT"]


def env():
    """Verify BOARD variables and construct exported variables"""

    if cij.ssh.env():
        cij.err("board.env: invalid SSH environment")
        return 1

    board = cij.env_to_dict(PREFIX, REQUIRED)   # Verify REQUIRED variables
    if board is None:
        cij.err("board.env: invalid BOARD environment")
        return 1

    board["CLASS"] = "_".join([board[r] for r in REQUIRED[:-1]])
    board["IDENT"] = "-".join([board["CLASS"], board["ALIAS"]])

    cij.env_export(PREFIX, EXPORTED, board)     # Export EXPORTED variables

    return 0



"""
    Environment for BLOCK devices
"""
import cij.util
import cij.ssh
import cij

PREFIX = "BLOCK"
REQUIRED = ["DEV_NAME"]
EXPORTED = ["DEV_PATH"]


def env():
    """Verify BLOCK variables and construct exported variables"""

    if cij.ssh.env():
        cij.err("cij.block.env: invalid SSH environment")
        return 1

    block = cij.env_to_dict(PREFIX, REQUIRED)

    block["DEV_PATH"] = "/dev/%s" % block["DEV_NAME"]

    cij.env_export(PREFIX, EXPORTED, block)

    return 0


def exists():
    """Verify that the ENV defined BLOCK device exists"""

    if env():
        cij.err("cij.block.exists: invalid BLOCK environment")
        return 1

    block = cij.env_to_dict(PREFIX, EXPORTED + REQUIRED)

    cmd = ['[[ -b "%s" ]]' % block["DEV_PATH"]]
    rcode, _, _ = cij.ssh.command(cmd, shell=True, echo=False)

    return rcode



"""
    Utilities for writing Python tests with CIJOE
"""
import sys
import os
import yaml
import cij.board
import cij.block
import cij.lnvm
import cij.util
import cij.ssh
import cij


PASS = 0
FAIL = 1
SKIP = 2
REQS = []


def tindex(spath=None):
    """
    Lists tindex in CIJ_TESTCASES

    @Returns On success, a list of filenames is returned. On error, None is
    returned
    """

    spath = spath if spath else os.environ.get("CIJ_TESTCASES", None)
    if spath is None:
        return None

    tests = []                          # Look for .sh files
    for root, _, files in os.walk(spath):
        if root != spath:
            continue

        tests += [f for f in files if f[-3:] in [".sh", ".py"]]

    return tests


def envs():
    """
    Return variables defined by modules required by test
    """

    variables = {}

    for req in REQS:
        prefix = req.upper()
        variables[prefix] = cij.env_to_dict(
            prefix, getattr(cij, req).REQUIRED + getattr(cij, req).EXPORTED
        )

    return variables


def require(req):
    """Add test requirement"""

    REQS.append(req)


def enter():
    """Enter the test, check requirements and setup aux. environment"""

    if cij.ssh.env():
        tfail("cij.test: invalid SSH environment")

    for req in REQS:
        if getattr(cij, req).env():
            tfail()

    cij.emph("cij.test: entering test")


def texit(msg=None, rcode=1):
    """Exit the test"""

    msg = ", msg: %r" % msg if msg else ""

    if rcode:
        cij.err("cij.test: FAILED%s" % msg)
    else:
        cij.good("cij.test: PASSED%s" % msg)

    sys.exit(rcode)


def tpass(msg=None):
    """Testing: Exit test indicating test passed"""

    texit(msg, 0)


def tfail(msg=None):
    """Testing: Exit test indicating test failed"""

    texit(msg, 1)


def command(cmd, ssh=True, shell=True, echo=True):
    """
    Execute the given 'cmd'

    @returns (rcode, stdout, stderr)
    """

    if ssh:
        return cij.ssh.command(cmd, shell, echo)

    return cij.util.execute(cmd, shell, echo)


def command_to_struct(cmd):
    """
    Same as `command` except it tries to convert stdout to struct

    @returns (rcode, struct, stderr, struct)
    """

    struct = None

    rcode, stdout, stderr = command(cmd)

    try:
        lines = []

        for line in stdout.splitlines():
            if line.strip().startswith("#"):
                continue

            lines.append(line)

        struct = yaml.safe_load("\n".join(lines))
    except (yaml.YAMLError) as exc:
        cij.err("could not parse stdout as yaml, exc: %r" % exc)

    return rcode, stdout, stderr, struct



#!/usr/bin/env python
# coding=utf-8
"""
xlsx.py      - Script providing operation of xls/xlsx

Class Basic:
    Basic.import_source()    - import data of basic performance
    Basic.generate_chart()   - generate chart of basic performance
"""
# pylint: disable=E0012,R0205
import os
from collections import Counter
import numpy as np
import xlsxwriter as xw
import matplotlib.pyplot as plt


class Basic(object):
    """
    Function:
        1. Create a excel file;
        2. Parse a data file (.txt) which has title on first line;
        3. Translate data into array format;
        4. Filter array according to input list, which element is dict;
        5. Specify x axis, y axis, data and series;
        6. Generate sheet with the coordinate figure
        7. Close and save excel file.
    Input:
        name: Target to generate. If file was exists, it would be regenerate.
    """
    def __init__(self, name):
        self.source_sheet = None
        self.source_data = None
        self.sheet_dict = {}

        if os.path.isfile(name):
            os.remove(name)

        self.workbook = xw.Workbook(name)

    def import_source(self, sheet, source, delimiter=","):
        """
        Function:
            Save original data into specific sheet, and try to translate data to float type
        Input:
            sheet: Must be a non exists sheet
            source: File path of source
        """
        # check input parameters
        if ' ' in sheet:
            raise RuntimeError("Error sheet name: %s" % sheet)

        if not source.endswith("txt") and not source.endswith("csv"):
            raise RuntimeError("Error source name: %s" % source)

        self.source_sheet = sheet

        source_data = np.loadtxt(source, dtype=str, delimiter=delimiter)
        self.source_data = {"title": source_data[0].tolist(),
                            "data": source_data[1:]}

        cell_format_title = self.workbook.add_format({'bold': True,
                                                      'font_name': u'等线',
                                                      'bg_color': '#c5d9f1',
                                                      'rotation': 45})
        cell_format = self.workbook.add_format({'bold': False,
                                                'font_name': u'等线',
                                                'num_format': 0})

        worksheet = self.workbook.add_worksheet(sheet)
        worksheet.write_row('A1', self.source_data['title'], cell_format_title)

        _, col_num = self.source_data['data'].shape
        for i in range(col_num):
            try:
                data_array = self.source_data['data'][:, i].astype(float)
            except ValueError:
                data_array = self.source_data['data'][:, i]

            worksheet.write_column(1, i, data_array.tolist(), cell_format)

    def generate_chart(self, properties):
        """
        Function:
            Generate and save chart to specific sheet.
        Input:
            sheet: If already exists, new chart will be added below.
                   Otherwise, it would create a new sheet;
            x_axis: Specify x axis;
            y_axis: Specify y axis;
            series: Specify series;
            filters: dict type, use to filter useful data from original data;
            title: if None, the chart will create without title;
            x_axis_name: if None, use x_axis instead;
            y_axis_name: if None, use y_axis instead;
        """
        # check input parameters
        if not {'x_axis', 'y_axis', 'series', 'filters'}.issubset(set(properties.keys())):
            raise RuntimeError("Error properties: %s" % properties.keys())

        # generate chart
        mask = self.__filter_data(properties['filters'])
        chart = self.__generate_chart(mask, properties)
        sheet = properties['sheet']

        # Add work sheet
        if sheet in self.sheet_dict.keys():
            self.sheet_dict[sheet] += 1
            worksheet = self.workbook.get_worksheet_by_name(sheet)
        else:
            self.sheet_dict[sheet] = 1
            worksheet = self.workbook.add_worksheet(sheet)
        worksheet.insert_chart('B%d' % (5 + (self.sheet_dict[sheet] - 1) * 35), chart)

    def __filter_data(self, filters):
        title = self.source_data["title"]
        data = self.source_data["data"]
        row_num, _ = self.source_data['data'].shape
        mask = np.array([True] * row_num)
        for key in filters.keys():
            target = filters.get(key)
            index = title.index(key)
            mask = mask & (data[:, index] == str(target))
        return mask

    def __generate_chart(self, mask, properties):
        chart = self.workbook.add_chart({'type': 'line'})

        # Title of chart
        if 'title' in properties:
            chart.set_title({"name": properties['title']})

        # Add series to chart
        series_index = self.source_data['title'].index(properties['series'])
        series_array = np.unique(self.source_data['data'][:, series_index])
        for series_name in series_array:
            mask_serials = mask & (self.source_data['data'][:, series_index] == series_name)
            x_axis_index = self.source_data['title'].index(properties['x_axis'])
            y_axis_index = self.source_data['title'].index(properties['y_axis'])
            x_data = self.__generate_data(x_axis_index, mask_serials)
            y_data = self.__generate_data(y_axis_index, mask_serials)
            chart.add_series({'categories': x_data,
                              'values': y_data,
                              'marker': {'type': 'diamond'},
                              'name': '%s' % series_name})

        # Set properties of x/y axis
        if 'x_axis_name' not in properties:
            properties['x_axis_name'] = properties['x_axis']
        if 'y_axis_name' not in properties:
            properties['y_axis_name'] = properties['y_axis']

        chart.set_x_axis({'name': properties['x_axis_name'],
                          'name_font': {'size': 14, 'bold': False},
                          'num_font': {'italic': True}})
        chart.set_y_axis({'name': properties['y_axis_name'],
                          'name_font': {'size': 14, 'bold': False},
                          'num_font': {'italic': True}})
        chart.set_size({"x_scale": 2,
                        "y_scale": 2})

        return chart

    def __generate_data(self, x_axis_index, mask):
        if 0 <= x_axis_index <= 25:
            col_num = chr(65 + x_axis_index)
        else:
            col_num = chr(65 + x_axis_index / 26 - 1) + chr(65 + x_axis_index % 26)

        data_list = []
        for index, item in enumerate(mask):
            if item:
                data_list.append("%s!$%s$%s" % (self.source_sheet, col_num, index + 2))

        data = "=(%s)" % ",".join(data_list)

        return data

    def close(self):
        """Close work book"""
        self.workbook.close()


class Gc(object):
    """
    Function:
        1. Create a GC excel file;
        2. Parse a data file (.txt) which has title on first line;
        3. Translate data into array format;
        4. Filter array according to input list, which element is dict;
        5. Specify x axis, y axis, data and series;
        6. Generate sheet with the coordinate figure
        7. Close and save excel file.
    Input:
        name: Target to generate. If file was exists, it would be regenerate.
    """
    def __init__(self, name):
        self.dst = name
        self.sheetname_dict = {}
        self.pic_list = []
        if os.path.isfile(name):
            os.remove(name)
        self.workbook = xw.Workbook(name)

    def gen_config_sheet(self, sheetname, plist):
        """ generate configuration"""
        worksheet_cfg = self.workbook.add_worksheet(sheetname)
        cell_format = self.workbook.add_format({'bold': False, 'font_name': u'等线'})
        cell_format_title = self.workbook.add_format({'border': 0, 'align': 'center',
                                                      'bg_color': '#c5d9f1', 'font_size': 12,
                                                      'font_name': u'等线', 'bold': False})
        worksheet_cfg.set_column('A:E', 40, cell_format)
        worksheet_cfg.write_row('A1', plist[0], cell_format_title)
        for i in range(1, len(plist)):
            worksheet_cfg.write_row('A%d' % (i+1), plist[i], cell_format)

    def gen_data_sheet(self, datafile, para_meter, scale=1.75, steady_time=300):
        """
        datafile, sheetname, x_axis_name, y_axis_name, title,
        Function:
            Turn realtime bw data into picture, and save into specific sheet
        Input:
            sheetname: If already exists, new chart will be added continuely.
                       Otherwise, it would create new sheet;
            x_axis_name: x_axis name;
            y_axis_name: y_axis name;
            title: picure name;
            scale； size of picture.
        """
        filename = os.path.splitext(os.path.split(datafile)[1])[0][:-5]
        para_meter['filename'] = filename
        source_data = np.loadtxt(datafile, dtype=int, delimiter=',')[:, :3]
        is_bw = 'bw'in para_meter['title'].lower()
        file_data = process_rt_data(source_data, is_bw)
        pic_path = generate_rt_pic(file_data, para_meter, scale)
        pic_path_steady = generate_steady_rt_pic(file_data, para_meter, scale, steady_time)
        if para_meter['sheetname'] in self.sheetname_dict.keys():
            self.sheetname_dict[para_meter['sheetname']] = \
                self.sheetname_dict[para_meter['sheetname']] + 1
            chart_sheet = self.workbook.get_worksheet_by_name(para_meter['sheetname'])
        else:
            self.sheetname_dict[para_meter['sheetname']] = 1
            chart_sheet = self.workbook.add_worksheet(para_meter['sheetname'])
        chart_sheet.insert_image('B%d' %
                                 (5 + (self.sheetname_dict[para_meter['sheetname']] - 1) * 30),
                                 pic_path)
        chart_sheet.insert_image('Q%d' %
                                 (5 + (self.sheetname_dict[para_meter['sheetname']] - 1) * 30),
                                 pic_path_steady)
        self.__insert_value(chart_sheet, file_data,
                            5 + (self.sheetname_dict[para_meter['sheetname']] - 1) * 30,
                            steady_time)
        self.pic_list.append(pic_path)
        self.pic_list.append(pic_path_steady)

    def __insert_value(self, chart_sheet, data, start_row, steady_time):
        start_row = start_row + 23
        chart_sheet.set_column('Q:T', 20)
        cell_format_title = self.workbook.add_format({'border': 1, 'bold': 1,
                                                      'align': 'center', 'bg_color': '#c5d9f1',
                                                      'font_size': 12, 'font_name': u'等线'})
        cell_format = self.workbook.add_format({'border': 1, 'bold': 1, 'font_name': u'等线'})
        list_value = ["Serials", "max_value", "min_value", "avg_value"]
        chart_sheet.write_row('Q%d' % start_row, list_value, cell_format_title)
        for key in data.keys():
            if len(data[key]) < steady_time:
                steady_time = len(data[key])
            start_row = start_row + len(data.keys())
            steady_value = data[key][-1 * steady_time:]
            list_value = ["Steady State %d" % key]
            max_value = np.max(steady_value[:, 1])
            list_value.append(max_value)
            min_value = np.min(steady_value[:, 1])
            list_value.append(min_value)
            mean_value = np.mean(steady_value[:, 1])
            list_value.append(mean_value)
            list_value_offset = ["Offset %d" % key,
                                 "%.2f%%" % float((max_value - mean_value) / mean_value * 100),
                                 "%.2f%%" % float((mean_value - min_value) / mean_value * 100), '']
            chart_sheet.write_row('Q%d' % start_row, list_value, cell_format)
            chart_sheet.write_row('Q%d' % (start_row + 1), list_value_offset, cell_format)

    def closexls(self):
        """ close the xlsx"""
        self.workbook.close()


def generate_rt_pic(process_data, para_meter, scale):
    """ generate rater pic"""
    pic_path = para_meter['filename'] + '.png'
    plt.figure(figsize=(5.6 * scale, 3.2 * scale))
    for key in process_data.keys():
        plt.plot(process_data[key][:, 0], process_data[key][:, 1], label=str(key))
    plt.title(para_meter['title'])
    plt.xlabel(para_meter['x_axis_name'])
    plt.ylabel(para_meter['y_axis_name'])
    plt.legend(loc='upper left')
    plt.savefig(pic_path)
    return pic_path


def generate_steady_rt_pic(process_data, para_meter, scale, steady_time):
    """ generate rate steady"""
    pic_path_steady = para_meter['filename'] + '_steady.png'
    plt.figure(figsize=(4 * scale, 2.5 * scale))
    for key in process_data.keys():
        if len(process_data[key]) < steady_time:
            steady_time = len(process_data[key])
        plt.scatter(process_data[key][-1 * steady_time:, 0],
                    process_data[key][-1 * steady_time:, 1], label=str(key), s=10)
        steady_value = np.mean(process_data[key][-1 * steady_time:, 1])
        steady_value_5 = steady_value * (1 + 0.05)
        steady_value_10 = steady_value * (1 + 0.1)
        steady_value_ng_5 = steady_value * (1 - 0.05)
        steady_value_ng_10 = steady_value * (1 - 0.1)
        plt.plot(process_data[key][-1 * steady_time:, 0], [steady_value] * steady_time, 'b')
        plt.plot(process_data[key][-1 * steady_time:, 0], [steady_value_5] * steady_time, 'g')
        plt.plot(process_data[key][-1 * steady_time:, 0],
                 [steady_value_ng_5] * steady_time, 'g')
        plt.plot(process_data[key][-1 * steady_time:, 0], [steady_value_10] * steady_time, 'r')
        plt.plot(process_data[key][-1 * steady_time:, 0],
                 [steady_value_ng_10] * steady_time, 'r')
    plt.title(para_meter['title'] + '(steady)')
    plt.xlabel(para_meter['x_axis_name'] + '(steady)')
    plt.ylabel(para_meter['y_axis_name'] + '(steady)')
    plt.legend(loc='upper left')
    plt.savefig(pic_path_steady)
    return pic_path_steady


def process_rt_data(source_data, is_bw=False):
    """ process data"""
    print("source_data length:", len(source_data))
    filter_data = {}
    for index in range(2):

        filter_mask = source_data[:, 2] == index

        if np.any(filter_mask):
            filter_data[index] = sum_data(round_data(source_data[filter_mask]), is_bw)
    return filter_data


def sum_data(filter_data, is_bw):
    """ caculate sum"""
    for index in range(len(filter_data) - 1):
        if filter_data[index][0] > filter_data[index + 1][0]:
            max_index = index + 1
            break
    else:
        max_index = len(filter_data)
    print("max_index: ", max_index + 1)
    num_jobs = int(round(len(filter_data) * 1.0 / max_index))
    print("num_jobs: ", num_jobs)

    dict_time = Counter(filter_data[:, 0])
    list_sum = []
    for time_index in range(1, max_index + 1):

        if dict_time.get(time_index * 1000, 0) != num_jobs:
            print("[WARNING] Time %d, number of data %d != num_jobs %d" % (
                time_index * 1000, dict_time.get(time_index * 1000, 0), num_jobs
            ))
            continue
        filter_mask = (filter_data[:, 0] == time_index * 1000)

        sum_rst = np.sum(filter_data[filter_mask][:, 1])
        if is_bw:
            sum_rst = sum_rst / 1024
        list_sum.append([time_index, sum_rst])
    return np.array(list_sum)


def round_data(filter_data):
    """ round the data"""
    for index, _ in enumerate(filter_data):
        filter_data[index][0] = round(filter_data[index][0] / 100.0) * 100.0
    return filter_data



"""
fio.py    - Script providing operation of fio

Classes:
    Job.import_parms()  - Import FIO parameters from dict
    Job.set_parm()      - Set FIO parameter
    Job.get_parm()      - Get FIO parameter
    Job.start()         - Start FIO in thread
    Job.join()          - Wait until FIO thread ended
    Job.result()        - Get result of FIO thread
    Job.run()           - Start FIO in foreground and return result

Require:
    cij.ssh.env()
"""
# pylint: disable=E0012,R0205
from collections import OrderedDict
from threading import Thread
import cij.ssh
import cij


def env():
    """Verify FIO variables and construct exported variables"""

    if cij.ssh.env():
        cij.err("cij.fio.env: invalid SSH environment")
        return 1
    return 0


def pkill():
    """Kill all of FIO processes"""

    if env():
        return 1

    cmd = ["ps -aux | grep fio | grep -v grep"]
    status, _, _ = cij.ssh.command(cmd, shell=True, echo=False)
    if not status:
        status, _, _ = cij.ssh.command(["pkill -f fio"], shell=True)
        if status:
            return 1
    return 0


class Threads(Thread):
    """Class of threads"""

    def __init__(self, target=None, args=()):
        Thread.__init__(self)
        self.target = target
        self.args = args
        self.output = None

    def run(self):
        """Start run thread"""

        self.output = self.target(*self.args)

    def result(self):
        """Get result of thread"""

        return self.output


class Job(object):
    """Class of FIO job"""

    def __init__(self):
        self.__thread = None
        self.__parm = OrderedDict()

    def __parse_parms(self):
        """Translate dict parameters to string"""

        args = list()
        for key, val in self.__parm.items():
            key = key.replace("FIO_", "").lower()

            if key == "runtime":
                args.append("--time_based")

            if val is None:
                args.append("--%s" % key)
            else:
                args.append("--%s=%s" % (key, val))
        return args

    def import_parms(self, args):
        """Import external dict to internal dict"""

        for key, val in args.items():
            self.set_parm(key, val)

    def set_parm(self, key, val=None):
        """
        Set parameter of FIO
        If val is None, the parameter is "--key"
        If val is not None, the parameter is "--key=val"
        """

        self.__parm[key] = val

    def get_parm(self, key):
        """Get parameter of FIO"""

        if key in self.__parm.keys():
            return self.__parm[key]

        return None

    def start(self):
        """Run FIO job in thread"""

        self.__thread = Threads(target=self.run, args=(True, True, False))
        self.__thread.setDaemon(True)
        self.__thread.start()

    def join(self, timeout=None):
        """Wait until the FIO thread terminates"""

        if self.__thread:
            self.__thread.join(timeout)

    def result(self):
        """Get result of FIO thread"""

        if self.__thread:
            return self.__thread.result()

        return None

    def run(self, shell=True, cmdline=False, echo=True):
        """Run FIO job"""

        if env():
            return 1

        cmd = ["fio"] + self.__parse_parms()
        if cmdline:
            cij.emph("cij.fio.run: shell: %r, cmd: %r" % (shell, cmd))

        return cij.ssh.command(cmd, shell, echo)



"""
    Miscellaneous utilities
"""
from __future__ import print_function
from subprocess import Popen, PIPE
import os
import re
import cij


def expand_path(path):
    """Expands variables from the given path and turns it into absolute path"""

    return os.path.abspath(os.path.expanduser(os.path.expandvars(path)))


def regex_find(pattern, content):
    """Find the given 'pattern' in 'content'"""

    find = re.findall(pattern, content)
    if not find:
        cij.err("pattern <%r> is invalid, no matches!" % pattern)
        cij.err("content: %r" % content)
        return ''

    if len(find) >= 2:
        cij.err("pattern <%r> is too simple, matched more than 2!" % pattern)
        cij.err("content: %r" % content)
        return ''

    return find[0]


def execute(cmd=None, shell=True, echo=True):
    """
    Execute the given 'cmd'

    @returns (rcode, stdout, stderr)
    """
    if echo:
        cij.emph("cij.util.execute: shell: %r, cmd: %r" % (shell, cmd))

    rcode = 1
    stdout, stderr = ("", "")

    if cmd:
        if shell:
            cmd = " ".join(cmd)

        proc = Popen(cmd, stdout=PIPE, stderr=PIPE, shell=shell, close_fds=True)
        stdout, stderr = proc.communicate()
        rcode = proc.returncode

    if rcode and echo:
        cij.warn("cij.util.execute: stdout: %s" % stdout)
        cij.err("cij.util.execute: stderr: %s" % stderr)
        cij.err("cij.util.execute: rcode: %s" % rcode)

    return rcode, stdout, stderr



"""
    Environment for LNVM
"""
import cij.util
import cij.ssh
import cij

PREFIX = "LNVM"
REQUIRED = ["BGN", "END", "DEV_TYPE"]
EXPORTED = ["DEV_NAME", "DEV_PATH"]


def env():
    """Verify LNVM variables and construct exported variables"""

    if cij.ssh.env():
        cij.err("cij.lnvm.env: invalid SSH environment")
        return 1

    lnvm = cij.env_to_dict(PREFIX, REQUIRED)
    nvme = cij.env_to_dict("NVME", ["DEV_NAME"])

    if "BGN" not in lnvm.keys():
        cij.err("cij.lnvm.env: invalid LNVM_BGN")
        return 1
    if "END" not in lnvm.keys():
        cij.err("cij.lnvm.env: invalid LNVM_END")
        return 1
    if "DEV_TYPE" not in lnvm.keys():
        cij.err("cij.lnvm.env: invalid LNVM_DEV_TYPE")
        return 1

    lnvm["DEV_NAME"] = "%sb%03de%03d" % (nvme["DEV_NAME"], int(lnvm["BGN"]), int(lnvm["END"]))
    lnvm["DEV_PATH"] = "/dev/%s" % lnvm["DEV_NAME"]

    cij.env_export(PREFIX, EXPORTED, lnvm)

    return 0


def create():
    """Create LNVM device"""

    if env():
        cij.err("cij.lnvm.create: Invalid LNVM ENV")
        return 1

    nvme = cij.env_to_dict("NVME", ["DEV_NAME"])
    lnvm = cij.env_to_dict(PREFIX, EXPORTED + REQUIRED)
    cij.emph("lnvm.create: LNVM_DEV_NAME: %s" % lnvm["DEV_NAME"])

    cmd = ["nvme lnvm create -d %s -n %s -t %s -b %s -e %s -f" % (
        nvme["DEV_NAME"], lnvm["DEV_NAME"], lnvm["DEV_TYPE"], lnvm["BGN"], lnvm["END"])]
    rcode, _, _ = cij.ssh.command(cmd, shell=True)
    if rcode:
        cij.err("cij.lnvm.create: FAILED")
        return 1

    return 0


def recover():
    """Recover LNVM device"""

    if env():
        cij.err("cij.lnvm.create: Invalid LNVM ENV")
        return 1

    nvme = cij.env_to_dict("NVME", ["DEV_NAME"])
    lnvm = cij.env_to_dict(PREFIX, EXPORTED + REQUIRED)
    cij.emph("lnvm.recover: LNVM_DEV_NAME: %s" % lnvm["DEV_NAME"])

    cmd = ["nvme lnvm create -d %s -n %s -t %s -b %s -e %s" % (
        nvme["DEV_NAME"], lnvm["DEV_NAME"], lnvm["DEV_TYPE"], lnvm["BGN"], lnvm["END"])]
    rcode, _, _ = cij.ssh.command(cmd, shell=True)
    if rcode:
        cij.err("cij.lnvm.recover: FAILED")
        return 1

    return 0


def remove():
    """Remove LNVM device"""

    if env():
        cij.err("cij.lnvm.create: Invalid LNVM ENV")
        return 1

    lnvm = cij.env_to_dict(PREFIX, EXPORTED + REQUIRED)
    cij.emph("lnvm.remove: LNVM_DEV_NAME: %s" % lnvm["DEV_NAME"])

    cmd = ["nvme lnvm remove -n %s" % (lnvm["DEV_NAME"])]
    rcode, _, _ = cij.ssh.command(cmd, shell=True)
    if rcode:
        cij.err("cij.lnvm.remove: FAILED")
        return 1

    return 0


def exists():
    """Verify that the ENV defined LNVM device exists"""

    if env():
        cij.err("cij.nvme.exists: Invalid NVMe ENV.")
        return 1

    lnvm = cij.env_to_dict(PREFIX, EXPORTED + REQUIRED)

    cmd = ['[[ -b "%s" ]]' % lnvm["DEV_PATH"]]
    rcode, _, _ = cij.ssh.command(cmd, shell=True, echo=False)
    if rcode:
        return False

    return True



"""
    Environment for DMESG
"""
# pylint: disable=E0012,R0205
from threading import Thread
import cij.ssh
import cij.util
import cij


def env():
    """Verify FIO variables and construct exported variables"""

    if cij.ssh.env():
        cij.err("cij.dmesg.env: invalid SSH environment")
        return 1
    return 0


class Job(object):
    """Class of DMESG job"""

    def __init__(self, fout):
        self.__thread = None
        self.__prefix = ["dmesg", "-w", "-T"]
        self.__suffix = [">>", fout]

    def __run(self, shell=True, echo=True):
        """Run DMESG job"""

        if env():
            return 1

        cij.emph("cij.dmesg.start: shell: %r, cmd: %r" % (shell, self.__prefix + self.__suffix))

        return cij.ssh.command(self.__prefix, shell, echo, self.__suffix)

    def start(self):
        """Start DMESG job in thread"""

        self.__thread = Thread(target=self.__run, args=(True, False))
        self.__thread.setDaemon(True)
        self.__thread.start()

    def terminate(self):
        """Terminate DMESG job"""

        if self.__thread:
            cmd = ["who am i"]
            status, output, _ = cij.util.execute(cmd, shell=True, echo=True)
            if status:
                cij.warn("cij.dmesg.terminate: who am i failed")
                return 1

            tty = output.split()[1]

            cmd = ["pkill -f '{}' -t '{}'".format(" ".join(self.__prefix), tty)]
            status, _, _ = cij.util.execute(cmd, shell=True, echo=True)
            if status:
                cij.warn("cij.dmesg.terminate: pkill failed")
                return 1

            self.__thread.join()
            self.__thread = None

        return 0



"""
ssh.py      - Script providing operation of SSH

Functions:
    ssh.env()       - Check environment of SSH connection
    ssh.command()   - Send SSH command to TARGET
    ssh.push()      - Push file to TARGET by SSH
    ssh.pull()      - Pull file from TARGET by SSH
    ssh.wait()      - Wait for TARGET get ready until timeout
    ssh.reboot()    - Reboot TARGET and wait for TARGET get ready

Require:
    SSH_KEY         - SSH key of TARGET
    SSH_PORT        - SSH port of TARGET
    SSH_USER        - SSH user of TARGET
    SSH_HOST        - Name or IP of TARGET
    SSH_CMD_TIME    - Measure time of SSH command
    SSH_CMD_TIMEOUT - Timeout for SSH command
"""
import time
import cij.util
import cij

PREFIX = "SSH"
REQUIRED = ["USER", "HOST"]
EXPORTED = []

DEFAULTS = {"PORT": "22", "CMD_TIME": "1"}
OPTIONAL = ["CMD_TIMEOUT", "KEY"]


def env():
    """Verify SSH variables and construct exported variables"""

    ssh = cij.env_to_dict(PREFIX, REQUIRED)
    if "KEY" in ssh:
        ssh["KEY"] = cij.util.expand_path(ssh["KEY"])

    if cij.ENV.get("SSH_PORT") is None:
        cij.ENV["SSH_PORT"] = "22"
        cij.warn("cij.ssh.env: SSH_PORT was not set, assigned: %r" % (
            cij.ENV.get("SSH_PORT")
        ))

    if cij.ENV.get("SSH_CMD_TIME") is None:
        cij.ENV["SSH_CMD_TIME"] = "1"
        cij.warn("cij.ssh.env: SSH_CMD_TIME was not set, assigned: %r" % (
            cij.ENV.get("SSH_CMD_TIME")
        ))

    return 0


def command(cmd, shell=True, echo=True, suffix=None):
    """SSH: Run the given command over SSH as defined in environment"""

    if env():
        cij.err("cij.ssh.command: Invalid SSH environment")
        return 1

    prefix = []

    if cij.ENV.get("SSH_CMD_TIME") == "1":
        prefix.append("/usr/bin/time")

    if cij.ENV.get("SSH_CMD_TIMEOUT"):
        prefix.append("timeout")
        prefix.append(cij.ENV.get("SSH_CMD_TIMEOUT"))

    prefix.append("ssh")

    args = []

    if cij.ENV.get("SSH_KEY"):
        args.append("-i")
        args.append(cij.ENV.get("SSH_KEY"))

    if cij.ENV.get("SSH_PORT"):
        args.append("-p")
        args.append(cij.ENV.get("SSH_PORT"))

    args.append("@".join([cij.ENV.get("SSH_USER"), cij.ENV.get("SSH_HOST")]))

    wrapped = prefix + args + ["'%s'" % " ".join(cmd)]
    if suffix:
        wrapped += suffix

    return cij.util.execute(wrapped, shell, echo)


def push(src, dst, folder=False):
    """SSH: push data to remote linux"""

    if env():
        cij.err("cij.ssh.push: Invalid SSH environment")
        return 1

    args = []

    if cij.ENV.get("SSH_KEY"):
        args.append("-i")
        args.append(cij.ENV.get("SSH_KEY"))

    if cij.ENV.get("SSH_PORT"):
        args.append("-P")
        args.append(cij.ENV.get("SSH_PORT"))

    if folder:
        args.append("-r")

    target = "%s:%s" % ("@".join([cij.ENV.get("SSH_USER"), cij.ENV.get("SSH_HOST")]), dst)
    wrapped = ["scp", " ".join(args), src, target]

    return cij.util.execute(wrapped, shell=True, echo=True)


def pull(src, dst, folder=False):
    """SSH: pull data from remote linux"""

    if env():
        cij.err("cij.ssh.pull: Invalid SSH environment")
        return 1

    args = []

    if cij.ENV.get("SSH_KEY"):
        args.append("-i")
        args.append(cij.ENV.get("SSH_KEY"))

    if cij.ENV.get("SSH_PORT"):
        args.append("-P")
        args.append(cij.ENV.get("SSH_PORT"))

    if folder:
        args.append("-r")

    target = "%s:%s" % ("@".join([cij.ENV.get("SSH_USER"), cij.ENV.get("SSH_HOST")]), src)
    wrapped = ["scp", " ".join(args), target, dst]

    return cij.util.execute(wrapped, shell=True, echo=True)


def wait(timeout=300):
    """Wait util target connected"""

    if env():
        cij.err("cij.ssh.wait: Invalid SSH environment")
        return 1

    timeout_backup = cij.ENV.get("SSH_CMD_TIMEOUT")

    try:
        time_start = time.time()

        cij.ENV["SSH_CMD_TIMEOUT"] = "3"

        while True:
            time_current = time.time()
            if (time_current - time_start) > timeout:
                cij.err("cij.ssh.wait: Timeout")
                return 1

            status, _, _ = command(["exit"], shell=True, echo=False)
            if not status:
                break

        cij.info("cij.ssh.wait: Time elapsed: %d seconds" % (time_current - time_start))

    finally:
        if timeout_backup is None:
            del cij.ENV["SSH_CMD_TIMEOUT"]
        else:
            cij.ENV["SSH_CMD_TIMEOUT"] = timeout_backup

    return 0


def reboot(timeout=300, extra=""):
    """Reboot target"""

    if env():
        cij.err("cij.ssh.reboot: Invalid SSH environment")
        return 1

    timeout_backup = cij.ENV.get("SSH_CMD_TIMEOUT")

    try:
        time_start = time.time()
        status, last_uptime, _ = command(["/usr/bin/uptime -s"], shell=True, echo=False)
        if status:
            return 1

        cij.ENV["SSH_CMD_TIMEOUT"] = "3"
        cij.info("cij.ssh.reboot: Target: %s" % cij.ENV.get("SSH_HOST"))
        command(["reboot %s" % extra], shell=True, echo=False)

        while True:
            time_current = time.time()
            if (time_current - time_start) > timeout:
                cij.err("cij.ssh.reboot: Timeout")
                return 1

            status, current_uptime, _ = command(["/usr/bin/uptime -s"], shell=True, echo=False)
            if not status and current_uptime != last_uptime:
                break

        cij.info("cij.ssh.reboot: Time elapsed: %d seconds" % (time_current - time_start))

    finally:
        if timeout_backup is None:
            del cij.ENV["SSH_CMD_TIMEOUT"]
        else:
            cij.ENV["SSH_CMD_TIMEOUT"] = timeout_backup

    return 0



"""
    Collection of Python utilities for CIJOE and CIJOE testing
"""
from __future__ import print_function
import time
import sys
import os

VERSION_MAJOR = 0
VERSION_MINOR = 0
VERSION_PATCH = 23
VERSION = "%d.%d.%d" % (VERSION_MAJOR, VERSION_MINOR, VERSION_PATCH)

PR_EMPH_CC = "\033[0;36m"
PR_GOOD_CC = "\033[0;32m"
PR_WARN_CC = "\033[0;33m"
PR_ERR_CC = "\033[0;31m"
PR_NC = "\033[0m"

ENV = os.environ
CIJ_ECHO_TIME_STAMP = ENV.get("CIJ_ECHO_TIME_STAMP")

EXTS = {
    "TPLANS": [".plan"],
    "TSUITES": [".suite"],
    "TCASES": [".py", ".sh"],
    "HOOKS": [".py", ".sh"],
}


def index(search_path, ext=None):
    """@returns a set of filenames with extension 'ext' in 'search_path'"""

    if ext is None:
        ext = "TCASES"

    fnames = set([])
    for _, _, files in os.walk(search_path):
        for fname in files:
            if os.path.splitext(fname)[-1] in EXTS[ext]:
                fnames.add(fname)

    return fnames


def get_time_stamp():
    """Get time stampe if CIJ_ECHO_TIME_STAMP is 1"""

    if CIJ_ECHO_TIME_STAMP == "1":
        return time.strftime('[%Y-%m-%d %H:%M:%S] ', time.localtime(time.time()))

    return ""


def info(txt):
    """Print, emphasized 'neutral', the given 'txt' message"""

    print("%s# %s%s%s" % (PR_EMPH_CC, get_time_stamp(), txt, PR_NC))
    sys.stdout.flush()


def good(txt):
    """Print, emphasized 'good', the given 'txt' message"""

    print("%s# %s%s%s" % (PR_GOOD_CC, get_time_stamp(), txt, PR_NC))
    sys.stdout.flush()


def warn(txt):
    """Print, emphasized 'warning', the given 'txt' message"""

    print("%s# %s%s%s" % (PR_WARN_CC, get_time_stamp(), txt, PR_NC))
    sys.stdout.flush()


def err(txt):
    """Print, emphasized 'error', the given 'txt' message"""

    print("%s# %s%s%s" % (PR_ERR_CC, get_time_stamp(), txt, PR_NC))
    sys.stdout.flush()


def emph(txt, rval=None):
    """Print, emphasized based on rval"""

    if rval is None:    # rval is not specified, use 'neutral'
        info(txt)
    elif rval == 0:     # rval is 0, by convention, this is 'good'
        good(txt)
    else:               # any other value, considered 'bad'
        err(txt)

def paths_from_env(prefix=None, names=None):
    """Construct dict of paths from environment variables'"""


    def expand_path(path):
        """Expands variables in 'path' and turns it into absolute path"""

        return os.path.abspath(os.path.expanduser(os.path.expandvars(path)))


    if prefix is None:
        prefix = "CIJ"
    if names is None:
        names = [
            "ROOT", "ENVS", "TESTPLANS", "TESTCASES", "TESTSUITES", "MODULES",
            "HOOKS", "TEMPLATES"
        ]

    conf = {v: os.environ.get("_".join([prefix, v])) for v in names}

    for env in (e for e in conf.keys() if e[:len(prefix)] in names and conf[e]):
        conf[env] = expand_path(conf[env])
        if not os.path.exists(conf[env]):
            err("%s_%s: %r, does not exist" % (prefix, env, conf[env]))

    return conf

def env_to_dict(prefix, names):
    """
    Construct dict from environment variables named: PREFIX_NAME

    @returns dict of names
    """

    env = {}
    for name in names:
        env[name] = ENV.get("_".join([prefix, name]))
        if env[name] is None:
            return None

    return env


def env_export(prefix, exported, env):
    """
    Define the list of 'exported' variables with 'prefix' with values from 'env'
    """

    for exp in exported:
        ENV["_".join([prefix, exp])] = env[exp]



#!/usr/bin/env python
"""
    Library functions for cij_reporter
"""
import datetime
import glob
import os
import ansi2html
import jinja2
import cij.runner
import cij


def extract_hook_names(ent):
    """Extract hook names from the given entity"""

    hnames = []
    for hook in ent["hooks"]["enter"] + ent["hooks"]["exit"]:
        hname = os.path.basename(hook["fpath_orig"])
        hname = os.path.splitext(hname)[0]
        hname = hname.strip()
        hname = hname.replace("_enter", "")
        hname = hname.replace("_exit", "")
        if hname in hnames:
            continue

        hnames.append(hname)

    hnames.sort()

    return hnames


def tcase_comment(tcase):
    """
    Extract testcase comment section / testcase description

    @returns the testcase-comment from the tcase["fpath"] as a list of strings
    """

    src = open(tcase["fpath"]).read()
    if len(src) < 3:
        cij.err("rprtr::tcase_comment: invalid src, tcase: %r" % tcase["name"])
        return None

    ext = os.path.splitext(tcase["fpath"])[-1]
    if ext not in [".sh", ".py"]:
        cij.err("rprtr::tcase_comment: invalid ext: %r, tcase: %r" % (
            ext, tcase["name"]
        ))
        return None

    comment = []
    for line in src.splitlines()[2:]:
        if ext == ".sh" and not line.startswith("#"):
            break
        elif ext == ".py" and not '"""' in line:
            break

        comment.append(line)

    return comment


def tcase_parse_descr(tcase):
    """Parse descriptions from the the given tcase"""

    descr_short = "SHORT"
    descr_long = "LONG"

    try:
        comment = tcase_comment(tcase)
    except (IOError, OSError, ValueError) as exc:
        comment = []
        cij.err("tcase_parse_descr: failed: %r, tcase: %r" % (exc, tcase))

    comment = [l for l in comment if l.strip()]     # Remove empty lines

    for line_number, line in enumerate(comment):
        if line.startswith("#"):
            comment[line_number] = line[1:]

    if comment:
        descr_short = comment[0]

    if len(comment) > 1:
        descr_long = "\n".join(comment[1:])

    return descr_short, descr_long


def runlogs_to_html(run_root):
    """
    Returns content of the given 'fpath' with HTML annotations, currently simply
    a conversion of ANSI color codes to HTML elements
    """

    if not os.path.isdir(run_root):
        return "CANNOT_LOCATE_LOGFILES"

    hook_enter = []
    hook_exit = []
    tcase = []
    for fpath in glob.glob(os.sep.join([run_root, "*.log"])):
        if "exit" in fpath:
            hook_exit.append(fpath)
            continue

        if "hook" in fpath:
            hook_enter.append(fpath)
            continue

        tcase.append(fpath)

    content = ""
    for fpath in hook_enter + tcase + hook_exit:
        content += "# BEGIN: run-log from log_fpath: %s\n" % fpath
        content += open(fpath, "r").read()
        content += "# END: run-log from log_fpath: %s\n\n" % fpath

    return content


def src_to_html(fpath):
    """
    Returns content of the given 'fpath' with HTML annotations for syntax
    highlighting
    """

    if not os.path.exists(fpath):
        return "COULD-NOT-FIND-TESTCASE-SRC-AT-FPATH:%r" % fpath

    # NOTE: Do SYNTAX highlight?

    return open(fpath, "r").read()


def aux_listing(aux_root):
    """Listing"""

    listing = []

    for root, _, fnames in os.walk(aux_root):
        count = len(aux_root.split(os.sep))
        prefix = root.split(os.sep)[count:]

        for fname in fnames:
            listing.append(os.sep.join(prefix + [fname]))

    return listing


def process_tsuite(tsuite):
    """Goes through the tsuite and processes "*.log" """

    # scoop of output from all run-logs

    tsuite["log_content"] = runlogs_to_html(tsuite["res_root"])
    tsuite["aux_list"] = aux_listing(tsuite["aux_root"])
    tsuite["hnames"] = extract_hook_names(tsuite)

    return True


def process_tcase(tcase):
    """Goes through the trun and processes "run.log" """

    tcase["src_content"] = src_to_html(tcase["fpath"])
    tcase["log_content"] = runlogs_to_html(tcase["res_root"])
    tcase["aux_list"] = aux_listing(tcase["aux_root"])
    tcase["descr_short"], tcase["descr_long"] = tcase_parse_descr(tcase)
    tcase["hnames"] = extract_hook_names(tcase)

    return True


def process_trun(trun):
    """Goes through the trun and processes "run.log" """

    trun["log_content"] = runlogs_to_html(trun["res_root"])
    trun["aux_list"] = aux_listing(trun["aux_root"])
    trun["hnames"] = extract_hook_names(trun)

    return True


def postprocess(trun):
    """Perform postprocessing of the given test run"""

    plog = []
    plog.append(("trun", process_trun(trun)))

    for tsuite in trun["testsuites"]:
        plog.append(("tsuite", process_tsuite(tsuite)))

        for tcase in tsuite["testcases"]:
            plog.append(("tcase", process_tcase(tcase)))

    for task, success in plog:
        if not success:
            cij.err("rprtr::postprocess: FAILED for %r" % task)

    return sum((success for task, success in plog))


def dset_to_html(dset, tmpl_fpath):
    """
    @returns A HTML representation of the given 'dset' using the template at
    'tmpl_fpath'
    """

    def stamp_to_datetime(stamp):
        """Create a date object from timestamp"""

        return datetime.datetime.fromtimestamp(int(stamp))

    def strftime(dtime, fmt):
        """Create a date object from timestamp"""

        return dtime.strftime(fmt)

    def ansi_to_html(ansi):
        """Convert the given ANSI text to HTML"""

        conv = ansi2html.Ansi2HTMLConverter(
            scheme="solarized",
            inline=True
        )
        html = conv.convert(ansi, full=False)

        with open("/tmp/jazz.html", "w") as html_file:
            html_file.write(html)

        return html

    tmpl_dpath = os.path.dirname(tmpl_fpath)
    tmpl_fname = os.path.basename(tmpl_fpath)

    env = jinja2.Environment(
        autoescape=True,
        loader=jinja2.FileSystemLoader(tmpl_dpath)
    )
    env.filters['stamp_to_datetime'] = stamp_to_datetime
    env.filters['strftime'] = strftime
    env.filters['ansi_to_html'] = ansi_to_html

    tmpl = env.get_template(tmpl_fname)

    return tmpl.render(dset=dset)


def rehome(old, new, struct):
    """
    Replace all absolute paths to "re-home" it
    """

    if old == new:
        return

    if isinstance(struct, list):
        for item in struct:
            rehome(old, new, item)
    elif isinstance(struct, dict):
        for key, val in struct.iteritems():
            if isinstance(val, (dict, list)):
                rehome(old, new, val)
            elif "conf" in key:
                continue
            elif "orig" in key:
                continue
            elif "root" in key or "path" in key:
                struct[key] = struct[key].replace(old, new)


def main(args):
    """Main entry point"""

    trun = cij.runner.trun_from_file(args.trun_fpath)

    rehome(trun["conf"]["OUTPUT"], args.output, trun)

    postprocess(trun)

    cij.emph("main: reports are uses tmpl_fpath: %r" % args.tmpl_fpath)
    cij.emph("main: reports are here args.output: %r" % args.output)

    html_fpath = os.sep.join([args.output, "%s.html" % args.tmpl_name])
    cij.emph("html_fpath: %r" % html_fpath)
    try:                                    # Create and store HTML report
        with open(html_fpath, 'w') as html_file:
            html_file.write(dset_to_html(trun, args.tmpl_fpath))
    except (IOError, OSError, ValueError) as exc:
        import traceback
        traceback.print_exc()
        cij.err("rprtr:main: exc: %s" % exc)
        return 1

    return 0



"""
    library functions for the CIJOE test runner, `cij_runner`.
"""
from __future__ import print_function
from subprocess import Popen, STDOUT
import shutil
import copy
import time
import os
import yaml
import cij.test
import cij

HOOK_PATTERNS = {
    "enter": [
        "%s.sh",
        "%s_enter.sh",
        "%s_enter.py",
    ],
    "exit": [
        "%s_exit.sh",
        "%s_exit.py"
    ]
}

HOOK = {
    "evars": {},

    "name": None,
    "fname": None,
    "fpath": None,
    "fpath_orig": None,

    "res_root": None,
    "log_fpath": None,
    "rcode": None,
    "wallc": None,
}

TESTSUITE = {
    "ident": None,
    "name": None,
    "alias": None,
    "hooks": {
        "enter": [],
        "exit": []
    },
    "evars": {},

    "fpath": None,
    "fname": None,
    "res_root": None,
    "aux_root": None,
    "aux_list": [],

    "status": "UNKN",
    "wallc": None,

    "testcases": [],
    "hooks_pr_tcase": [],
}

TESTCASE = {
    "ident": None,
    "fpath": None,
    "fname": None,
    "name": None,
    "res_root": None,
    "aux_root": None,
    "aux_list": [],
    "log_fpath": None,

    "hooks": None,
    "evars": {},

    "status": "UNKN",
    "rcode": None,
    "wallc": None,
}

TRUN = {
    "ver": None,
    "conf": None,
    "evars": {},
    "progress": {
        "PASS": 0,
        "FAIL": 0,
        "UNKN": 0
    },
    "stamp": {
        "begin": None,
        "end": None
    },
    "hooks": {
        "enter": [],
        "exit": []
    },
    "res_root": None,
    "aux_root": None,
    "aux_list": [],

    "testsuites": [],

    "status": "UNKN",
    "wallc": None,
}


def yml_fpath(output_path):
    """Returns the path to the trun-file"""

    return os.sep.join([output_path, "trun.yml"])


def script_run(trun, script):
    """Execute a script or testcase"""

    if trun["conf"]["VERBOSE"]:
        cij.emph("rnr:script:run { script: %s }" % script)
        cij.emph("rnr:script:run:evars: %s" % script["evars"])

    launchers = {
        ".py": "python",
        ".sh": "source"
    }

    ext = os.path.splitext(script["fpath"])[-1]
    if not ext in launchers.keys():
        cij.err("rnr:script:run { invalid script[\"fpath\"]: %r }" % script["fpath"])
        return 1

    launch = launchers[ext]

    with open(script["log_fpath"], "a") as log_fd:
        log_fd.write("# script_fpath: %r\n" % script["fpath"])
        log_fd.flush()

        bgn = time.time()
        cmd = [
            'bash', '-c',
            'CIJ_ROOT=$(cij_root) && '
            'source $CIJ_ROOT/modules/cijoe.sh && '
            'source %s && '
            'CIJ_TEST_RES_ROOT="%s" %s %s ' % (
                trun["conf"]["ENV_FPATH"],
                script["res_root"],
                launch,
                script["fpath"]
            )
        ]
        if trun["conf"]["VERBOSE"] > 1:
            cij.emph("rnr:script:run { cmd: %r }" % " ".join(cmd))

        evars = os.environ.copy()
        evars.update({k: str(script["evars"][k]) for k in script["evars"]})

        process = Popen(
            cmd,
            stdout=log_fd,
            stderr=STDOUT,
            cwd=script["res_root"],
            env=evars
        )
        process.wait()

        script["rcode"] = process.returncode
        script["wallc"] = time.time() - bgn

    if trun["conf"]["VERBOSE"]:
        cij.emph("rnr:script:run { wallc: %02f }" % script["wallc"])
        cij.emph(
            "rnr:script:run { rcode: %r } " % script["rcode"],
            script["rcode"]
        )

    return script["rcode"]

def hook_setup(parent, hook_fpath):
    """Setup hook"""

    hook = copy.deepcopy(HOOK)
    hook["name"] = os.path.splitext(os.path.basename(hook_fpath))[0]
    hook["name"] = hook["name"].replace("_enter", "").replace("_exit", "")
    hook["res_root"] = parent["res_root"]
    hook["fpath_orig"] = hook_fpath
    hook["fname"] = "hook_%s" % os.path.basename(hook["fpath_orig"])
    hook["fpath"] = os.sep.join([hook["res_root"], hook["fname"]])
    hook["log_fpath"] = os.sep.join([
        hook["res_root"],
        "%s.log" % hook["fname"]
    ])

    hook["evars"].update(copy.deepcopy(parent["evars"]))

    shutil.copyfile(hook["fpath_orig"], hook["fpath"])

    return hook

def hooks_setup(trun, parent, hnames=None):
    """
    Setup test-hooks
    @returns dict of hook filepaths {"enter": [], "exit": []}
    """

    hooks = {
        "enter": [],
        "exit": []
    }

    if hnames is None:       # Nothing to do, just return the struct
        return hooks

    for hname in hnames:      # Fill out paths
        for med in HOOK_PATTERNS:
            for ptn in HOOK_PATTERNS[med]:
                fpath = os.sep.join([trun["conf"]["HOOKS"], ptn % hname])
                if not os.path.exists(fpath):
                    continue

                hook = hook_setup(parent, fpath)
                if not hook:
                    continue

                hooks[med].append(hook)

        if not hooks["enter"] + hooks["exit"]:
            cij.err("rnr:hooks_setup:FAIL { hname: %r has no files }" % hname)
            return None

    return hooks


def trun_to_file(trun, fpath=None):
    """Dump the given trun to file"""

    if fpath is None:
        fpath = yml_fpath(trun["conf"]["OUTPUT"])

    with open(fpath, 'w') as yml_file:
        data = yaml.dump(trun, explicit_start=True, default_flow_style=False)
        yml_file.write(data)


def trun_from_file(fpath):
    """Returns trun from the given fpath"""

    with open(fpath, 'r') as yml_file:
        return yaml.safe_load(yml_file)


def trun_emph(trun):
    """Print essential info on"""

    if trun["conf"]["VERBOSE"] > 1:               # Print environment variables
        cij.emph("rnr:CONF {")
        for cvar in sorted(trun["conf"].keys()):
            cij.emph("  % 16s: %r" % (cvar, trun["conf"][cvar]))
        cij.emph("}")

    if trun["conf"]["VERBOSE"]:
        cij.emph("rnr:INFO {")
        cij.emph("  OUTPUT: %r" % trun["conf"]["OUTPUT"])
        cij.emph("  yml_fpath: %r" % yml_fpath(trun["conf"]["OUTPUT"]))
        cij.emph("}")


def tcase_setup(trun, parent, tcase_fname):
    """
    Create and initialize a testcase
    """
    #pylint: disable=locally-disabled, unused-argument

    case = copy.deepcopy(TESTCASE)

    case["fname"] = tcase_fname
    case["fpath_orig"] = os.sep.join([trun["conf"]["TESTCASES"], case["fname"]])

    if not os.path.exists(case["fpath_orig"]):
        cij.err('rnr:tcase_setup: !case["fpath_orig"]: %r' % case["fpath_orig"])
        return None

    case["name"] = os.path.splitext(case["fname"])[0]
    case["ident"] = "/".join([parent["ident"], case["fname"]])

    case["res_root"] = os.sep.join([parent["res_root"], case["fname"]])
    case["aux_root"] = os.sep.join([case["res_root"], "_aux"])
    case["log_fpath"] = os.sep.join([case["res_root"], "run.log"])

    case["fpath"] = os.sep.join([case["res_root"], case["fname"]])

    case["evars"].update(copy.deepcopy(parent["evars"]))

    # Initalize
    os.makedirs(case["res_root"])                       # Create DIRS
    os.makedirs(case["aux_root"])
    shutil.copyfile(case["fpath_orig"], case["fpath"])  # Copy testcase

    # Initialize hooks
    case["hooks"] = hooks_setup(trun, case, parent.get("hooks_pr_tcase"))

    return case


def tsuite_exit(trun, tsuite):
    """Triggers when exiting the given testsuite"""

    if trun["conf"]["VERBOSE"]:
        cij.emph("rnr:tsuite:exit")

    rcode = 0
    for hook in reversed(tsuite["hooks"]["exit"]):      # EXIT-hooks
        rcode = script_run(trun, hook)
        if rcode:
            break

    if trun["conf"]["VERBOSE"]:
        cij.emph("rnr:tsuite:exit { rcode: %r } " % rcode, rcode)

    return rcode


def tsuite_enter(trun, tsuite):
    """Triggers when entering the given testsuite"""

    if trun["conf"]["VERBOSE"]:
        cij.emph("rnr:tsuite:enter { name: %r }" % tsuite["name"])

    rcode = 0
    for hook in tsuite["hooks"]["enter"]:     # ENTER-hooks
        rcode = script_run(trun, hook)
        if rcode:
            break

    if trun["conf"]["VERBOSE"]:
        cij.emph("rnr:tsuite:enter { rcode: %r } " % rcode, rcode)

    return rcode

def tsuite_setup(trun, declr, enum):
    """
    Creates and initialized a TESTSUITE struct and site-effects such as creating
    output directories and forwarding initialization of testcases
    """

    suite = copy.deepcopy(TESTSUITE)  # Setup the test-suite

    suite["name"] = declr.get("name")
    if suite["name"] is None:
        cij.err("rnr:tsuite_setup: no testsuite is given")
        return None

    suite["alias"] = declr.get("alias")
    suite["ident"] = "%s_%d" % (suite["name"], enum)

    suite["res_root"] = os.sep.join([trun["conf"]["OUTPUT"], suite["ident"]])
    suite["aux_root"] = os.sep.join([suite["res_root"], "_aux"])

    suite["evars"].update(copy.deepcopy(trun["evars"]))
    suite["evars"].update(copy.deepcopy(declr.get("evars", {})))

    # Initialize
    os.makedirs(suite["res_root"])
    os.makedirs(suite["aux_root"])

    # Setup testsuite-hooks
    suite["hooks"] = hooks_setup(trun, suite, declr.get("hooks"))

    # Forward from declaration
    suite["hooks_pr_tcase"] = declr.get("hooks_pr_tcase", [])

    suite["fname"] = "%s.suite" % suite["name"]
    suite["fpath"] = os.sep.join([trun["conf"]["TESTSUITES"], suite["fname"]])

    #
    # Load testcases from .suite file OR from declaration
    #
    tcase_fpaths = []                               # Load testcase fpaths
    if os.path.exists(suite["fpath"]):              # From suite-file
        suite_lines = (
            l.strip() for l in open(suite["fpath"]).read().splitlines()
        )
        tcase_fpaths.extend(
            (l for l in suite_lines if len(l) > 1 and l[0] != "#")
        )
    else:                                           # From declaration
        tcase_fpaths.extend(declr.get("testcases", []))

    # NOTE: fix duplicates; allow them
    # NOTE: Currently hot-fixed here
    if len(set(tcase_fpaths)) != len(tcase_fpaths):
        cij.err("rnr:suite: failed: duplicate tcase in suite not supported")
        return None

    for tcase_fname in tcase_fpaths:                # Setup testcases
        tcase = tcase_setup(trun, suite, tcase_fname)
        if not tcase:
            cij.err("rnr:suite: failed: tcase_setup")
            return None

        suite["testcases"].append(tcase)

    return suite


def tcase_exit(trun, tsuite, tcase):
    """..."""
    #pylint: disable=locally-disabled, unused-argument

    if trun["conf"]["VERBOSE"]:
        cij.emph("rnr:tcase:exit { fname: %r }" % tcase["fname"])

    rcode = 0
    for hook in reversed(tcase["hooks"]["exit"]):    # tcase EXIT-hooks
        rcode = script_run(trun, hook)
        if rcode:
            break

    if trun["conf"]["VERBOSE"]:
        cij.emph("rnr:tcase:exit { rcode: %r }" % rcode, rcode)

    return rcode


def tcase_enter(trun, tsuite, tcase):
    """
    setup res_root and aux_root, log info and run tcase-enter-hooks

    @returns 0 when all hooks succeed, some value othervise
    """
    #pylint: disable=locally-disabled, unused-argument

    if trun["conf"]["VERBOSE"]:
        cij.emph("rnr:tcase:enter")
        cij.emph("rnr:tcase:enter { fname: %r }" % tcase["fname"])
        cij.emph("rnr:tcase:enter { log_fpath: %r }" % tcase["log_fpath"])

    rcode = 0
    for hook in tcase["hooks"]["enter"]:    # tcase ENTER-hooks
        rcode = script_run(trun, hook)
        if rcode:
            break

    if trun["conf"]["VERBOSE"]:
        cij.emph("rnr:tcase:exit: { rcode: %r }" % rcode, rcode)

    return rcode

def trun_exit(trun):
    """Triggers when exiting the given testrun"""

    if trun["conf"]["VERBOSE"]:
        cij.emph("rnr:trun:exit")

    rcode = 0
    for hook in reversed(trun["hooks"]["exit"]):    # EXIT-hooks
        rcode = script_run(trun, hook)
        if rcode:
            break

    if trun["conf"]["VERBOSE"]:
        cij.emph("rnr:trun::exit { rcode: %r }" % rcode, rcode)

    return rcode


def trun_enter(trun):
    """Triggers when entering the given testrun"""

    if trun["conf"]["VERBOSE"]:
        cij.emph("rnr:trun::enter")

    trun["stamp"]["begin"] = int(time.time())     # Record start timestamp

    rcode = 0
    for hook in trun["hooks"]["enter"]:     # ENTER-hooks
        rcode = script_run(trun, hook)
        if rcode:
            break

    if trun["conf"]["VERBOSE"]:
        cij.emph("rnr:trun::enter { rcode: %r }" % rcode, rcode)

    return rcode


def trun_setup(conf):
    """
    Setup the testrunner data-structure, embedding the parsed environment
    variables and command-line arguments and continues with setup for testplans,
    testsuites, and testcases
    """

    declr = None
    try:
        with open(conf["TESTPLAN_FPATH"]) as declr_fd:
            declr = yaml.safe_load(declr_fd)
    except AttributeError as exc:
        cij.err("rnr: %r" % exc)

    if not declr:
        return None

    trun = copy.deepcopy(TRUN)
    trun["ver"] = cij.VERSION

    trun["conf"] = copy.deepcopy(conf)
    trun["res_root"] = conf["OUTPUT"]
    trun["aux_root"] = os.sep.join([trun["res_root"], "_aux"])
    trun["evars"].update(copy.deepcopy(declr.get("evars", {})))

    os.makedirs(trun["aux_root"])

    hook_names = declr.get("hooks", [])
    if "lock" not in hook_names:
        hook_names = ["lock"] + hook_names

    if hook_names[0] != "lock":
        return None

    # Setup top-level hooks
    trun["hooks"] = hooks_setup(trun, trun, hook_names)

    for enum, declr in enumerate(declr["testsuites"]):  # Setup testsuites
        tsuite = tsuite_setup(trun, declr, enum)
        if tsuite is None:
            cij.err("main::FAILED: setting up tsuite: %r" % tsuite)
            return 1

        trun["testsuites"].append(tsuite)
        trun["progress"]["UNKN"] += len(tsuite["testcases"])

    return trun


def main(conf):
    """CIJ Test Runner main entry point"""

    fpath = yml_fpath(conf["OUTPUT"])
    if os.path.exists(fpath):   # YAML exists, we exit, it might be RUNNING!
        cij.err("main:FAILED { fpath: %r }, exists" % fpath)
        return 1

    trun = trun_setup(conf)         # Construct 'trun' from 'conf'
    if not trun:
        return 1

    trun_to_file(trun)              # Persist trun
    trun_emph(trun)                 # Print trun before run

    tr_err = 0
    tr_ent_err = trun_enter(trun)
    for tsuite in (ts for ts in trun["testsuites"] if not tr_ent_err):

        ts_err = 0
        ts_ent_err = tsuite_enter(trun, tsuite)
        for tcase in (tc for tc in tsuite["testcases"] if not ts_ent_err):

            tc_err = tcase_enter(trun, tsuite, tcase)
            if not tc_err:
                tc_err += script_run(trun, tcase)
                tc_err += tcase_exit(trun, tsuite, tcase)

            tcase["status"] = "FAIL" if tc_err else "PASS"

            trun["progress"][tcase["status"]] += 1  # Update progress
            trun["progress"]["UNKN"] -= 1

            ts_err += tc_err                        # Accumulate errors

            trun_to_file(trun)                      # Persist trun

        if not ts_ent_err:
            ts_err += tsuite_exit(trun, tsuite)

        ts_err += ts_ent_err                        # Accumulate errors
        tr_err += ts_err

        tsuite["status"] = "FAIL" if ts_err else "PASS"

        cij.emph("rnr:tsuite %r" % tsuite["status"], tsuite["status"] != "PASS")

    if not tr_ent_err:
        trun_exit(trun)

    tr_err += tr_ent_err
    trun["status"] = "FAIL" if tr_err else "PASS"

    trun["stamp"]["end"] = int(time.time()) + 1         # END STAMP
    trun_to_file(trun)                                  # PERSIST

    cij.emph("rnr:main:progress %r" % trun["progress"])
    cij.emph("rnr:main:trun %r" % trun["status"], trun["status"] != "PASS")

    return trun["progress"]["UNKN"] + trun["progress"]["FAIL"]



"""
usb.py      - Script providing operation for USB-RELAY

Classes:
    Relay.power_on  - 230v power on
    Relay.power_off - 230v power off
    Relay.power_btn - Target's power button

Require:
    USB_RELAY           - USB-RELAY is connect to workstation or not
    USB_RELAY_POWER_ON  - USB-RELAY port of 230v power on
    USB_RELAY_POWER_OFF - USB-RELAY port of 230v power off
    USB_RELAY_POWER_BTN - USB-RELAY port of target's power button
"""
# pylint: disable=E0012,R0205
import os
import time
from collections import OrderedDict
import cij.util
import cij


class Relay(object):
    """USB-RELAY operation"""

    def __init__(self):
        if os.environ.get("USB_RELAY") != "1":
            raise RuntimeError("cij.usb.relay: Invalid USB_RELAY")

        self.__power_on_port = os.environ.get("USB_RELAY_POWER_ON")
        self.__power_off_port = os.environ.get("USB_RELAY_POWER_OFF")
        self.__power_btn_port = os.environ.get("USB_RELAY_POWER_BTN")

        self.__field = OrderedDict()
        self.__device = None

        if self.__refresh():
            raise RuntimeError("cij.usb.relay: Error refresh status")

        if self.__reset():
            raise RuntimeError("cij.usb.relay: Error reset status")

    def __refresh(self):
        status, stdout, _ = cij.util.execute(["usbrelay"])
        if status:
            cij.err("cij.usb.relay: Error run usbrelay")
            return 1

        for line in stdout.strip().split("\n"):
            name, value = line.split("=")
            self.__field[name] = value

            if self.__device is None:
                self.__device = name.split("_")[0]
                cij.info("Device: %s" % self.__device)

            cij.info("%s: %s" % (name, value))

        if self.__field is None:
            cij.err("cij.usb.relay: Error usbrelay field")
            return 1

        return 0

    def __reset(self):
        if self.__field is None:
            cij.err("cij.usb.relay: Error usbrelay field")
            return 1

        for name, value in self.__field.items():
            if value == "1":
                if self.__set(name, 0):
                    return 1
        return 0

    def __set(self, name, state):
        if name not in self.__field.keys():
            cij.err("cij.usb.relay: Unknown name: %s" % name)
            return 1

        status, _, _ = cij.util.execute(["usbrelay %s=%s" % (name, state)])
        if status:
            return 1
        return 0

    def __press(self, port, interval=200):
        name = "%s_%s" % (self.__device, port)

        if self.__set(name, 1):
            return 1

        time.sleep(float(interval) / 1000)

        if self.__set(name, 0):
            return 1

        return 0

    def power_on(self, interval=200):
        """230v power on"""
        if self.__power_on_port is None:
            cij.err("cij.usb.relay: Invalid USB_RELAY_POWER_ON")
            return 1

        return self.__press(self.__power_on_port, interval=interval)

    def power_off(self, interval=200):
        """230v power off"""
        if self.__power_off_port is None:
            cij.err("cij.usb.relay: Invalid USB_RELAY_POWER_OFF")
            return 1

        return self.__press(self.__power_off_port, interval=interval)

    def power_btn(self, interval=200):
        """TARGET power button"""
        if self.__power_btn_port is None:
            cij.err("cij.usb.relay: Invalid USB_RELAY_POWER_BTN")
            return 1

        return self.__press(self.__power_btn_port, interval=interval)



"""
    Environment for PCI devices
"""
import os
import cij.util
import cij.ssh
import cij

PREFIX = "PCI"
REQUIRED = ["DEV_NAME"]
EXPORTED = ["DEV_PATH"]


def env():
    """Verify PCI variables and construct exported variables"""

    if cij.ssh.env():
        cij.err("cij.pci.env: invalid SSH environment")
        return 1

    pci = cij.env_to_dict(PREFIX, REQUIRED)

    pci["BUS_PATH"] = "/sys/bus/pci"
    pci["DEV_PATH"] = os.sep.join([pci["BUS_PATH"], "devices", pci["DEV_NAME"]])

    cij.env_export(PREFIX, EXPORTED, pci)

    return 0


def exists():
    """Verify that the ENV defined PCI device exists"""

    if env():
        cij.err("cij.pci.exists: invalid PCI environment")
        return 1

    pci = cij.env_to_dict(PREFIX, EXPORTED + REQUIRED)

    cmd = ['[[ -b "%s" ]]' % pci["DEV_PATH"]]
    rcode, _, _ = cij.ssh.command(cmd, shell=True, echo=False)

    return rcode



"""
    Definition of CIJOE distribution package
"""
import codecs
import glob
import os
from setuptools import setup

def read(*parts):
    """Read parts to use a e.g. long_description"""

    here = os.path.abspath(os.path.dirname(__file__))

    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with codecs.open(os.path.join(here, *parts), 'r') as pfp:
        return pfp.read()

setup(
    name="cijoe",
    version="0.0.23",
    description="Tools for systems development and testing",
    long_description=read('README.rst'),
    author="Simon A. F. Lund",
    author_email="slund@cnexlabs.com",
    url="https://github.com/refenv/cijoe",
    license="Apache License 2.0",
    install_requires=[
        "ansi2html (>=1.5.2)", "jinja2 (>=2.0)", "pyyaml (>=3.10)",
        "xlsxwriter", "matplotlib", "kmdo"
    ],
    zip_safe=False,
    packages=["cij"],
    package_dir={"": "modules"},
    data_files=[
        ("bin", glob.glob("bin/*")),

        ("share/cijoe/hooks", glob.glob("hooks/*")),
        ("share/cijoe/modules", glob.glob("modules/*.sh")),
        ("share/cijoe/templates", glob.glob("templates/*")),

        ("share/cijoe/envs", glob.glob("envs/*")),

        ("share/cijoe/testfiles", glob.glob("testfiles/*")),
        ("share/cijoe/testcases", glob.glob("testcases/*")),
        ("share/cijoe/testsuites", glob.glob("testsuites/*")),
        ("share/cijoe/testplans", glob.glob("testplans/*"))
    ],
    options={'bdist_wheel':{'universal':True}},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Topic :: Text Processing",
        "Topic :: Utilities",
        "Topic :: Software Development",
        "Topic :: Software Development :: Testing"
    ],
)



#!/usr/bin/env python
"""
    Example of a testcase implemented as a Python script

    The test implementation itself just verifies whether it can execute a
    command without error, it is just a demonstration on how commands should be
    invoked and how test-status must be communicated
"""
import cij.test
import cij.ssh
import cij
cij.test.enter()

def main():
    """
    @returns cij.test.PASS on success and cij.test.FAIL otherwise
    """

    cmd = ["lspci"]
    rcode, _, _ = cij.ssh.command(cmd, shell=True, echo=False)

    return cij.test.FAIL if rcode else cij.test.PASS

if __name__ == "__main__":
    cij.test.texit(rcode=main())



"""
bin.py      - Script providing operation of buffer

Classes:
    Buffer.length()     - Get length of array
    Buffer.size()       - Get size of buffer
    Buffer.types()      - Get type of item
    Buffer.write()      - Write buffer to file
    Buffer.read()       - Read buffer from file
    Buffer.dump()       - Dump buffer item to console
    Buffer.compare()    - Compare two buffers

"""
# pylint: disable=E0012,R0205
from ctypes import c_uint8, memmove, sizeof, cast, POINTER, Structure, Union, Array
import cij


def dump(buf, indent=0, skip=""):
    """Dump UnionType/StructType to STDOUT"""
    if not isinstance(type(buf), (type(Union), type(Structure))):
        raise RuntimeError("Error type(%s)" % type(buf))

    for field in getattr(buf, '_fields_'):
        name, types = field[0], field[1]
        if name in skip:
            return
        value = getattr(buf, name)

        if isinstance(types, (type(Union), type(Structure))):
            cij.info("%s%s:" % (" " * indent, name))
            dump(value, indent+2, skip)
        elif isinstance(types, type(Array)):
            for i, item in enumerate(value):
                name_index = "%s[%s]" % (name, i)

                if isinstance(types, (type(Union), type(Structure))):
                    cij.info("%s%s:" % (" " * indent, name_index))
                    dump(item, indent + 2, skip)
                else:
                    cij.info("%s%-12s: 0x%x" % (" " * indent, name_index, item))
        else:
            cij.info("%s%-12s: 0x%x" % (" " * indent, name, value))


def compare(buf_a, buf_b, ignore):
    """Compare of two Buffer item"""
    for field in getattr(buf_a, '_fields_'):
        name, types = field[0], field[1]

        if name in ignore:
            continue

        val_a = getattr(buf_a, name)
        val_b = getattr(buf_b, name)

        if isinstance(types, (type(Union), type(Structure))):
            if compare(val_a, val_b, ignore):
                return 1
        elif isinstance(types, type(Array)):
            for i, _ in enumerate(val_a):
                if isinstance(types, (type(Union), type(Structure))):
                    if compare(val_a[i], val_b[i], ignore):
                        return 1
                else:
                    if val_a[i] != val_b[i]:
                        return 1
        else:
            if val_a != val_b:
                return 1

    return 0


class Buffer(object):
    """Ctypes operation"""

    def __init__(self, types=c_uint8, length=1):
        self.m_types = types
        self.m_len = length
        self.m_sizeof = sizeof(self.m_types)
        self.m_size = self.m_len * self.m_sizeof
        self.m_buf = (self.m_types * self.m_len)()

    def __getitem__(self, key):
        return self.m_buf[key]

    def __setitem__(self, key, val):
        self.m_buf[key] = val

    def length(self):
        """Get length of types"""

        return self.m_len

    def size(self):
        """Get size of buffer"""

        return self.m_size

    def types(self):
        """Get types of buffer"""

        return self.m_types

    def memcopy(self, stream, offset=0, length=float("inf")):
        """Copy stream to buffer"""
        data = [ord(i) for i in list(stream)]
        size = min(length, len(data), self.m_size)
        buff = cast(self.m_buf, POINTER(c_uint8))
        for i in range(size):
            buff[offset + i] = data[i]

    def write(self, path):
        """Write buffer to file"""

        with open(path, "wb") as fout:
            fout.write(self.m_buf)

    def read(self, path):
        """Read file to buffer"""

        with open(path, "rb") as fout:
            memmove(self.m_buf, fout.read(self.m_size), self.m_size)

    def dump(self, offset=0, length=1):
        """Dump item"""

        for i in range(offset, offset + length):
            if "ctypes" in str(self.m_types):
                cij.info("Buff[%s]: %s" % (i, self.m_buf[i]))
            else:
                cij.info("Buff[%s]:" % i)
                dump(self.m_buf[i], 2)

    def compare(self, buf, offset=0, length=1, ignore=""):
        """Compare buffer"""

        for i in range(offset, offset + length):
            if isinstance(self.m_types, (type(Union), type(Structure))):
                if compare(self.m_buf[i], buf[i], ignore=ignore):
                    return 1
            elif self.m_buf[i] != buf[i]:
                return 1

        return 0



"""
    ipmi.py - Script providing convenience functions for invoking ipmi

    Functions:

    ipmi.env        - Sets default vars for IPMI wrapping
    ipmi.cmd <CMD>  - Execute ipmitool command <CMD>
    ipmi.pwr_on     - Power on system
    ipmi.pwr_off    - Power off system
    ipmi.pwr_reset  - Power reset system

    Variables:

    IPMI_USER - login on server
    IPMI_PASS - password to server
    IPMI_HOST - server host
    IPMI_PORT - server port
"""
import cij.util
import cij

PREFIX = "IMPI"
REQUIRED = ["USER", "PASS", "HOST", "PORT"]
EXPORTED = []


def env():
    """Verify IPMI environment"""

    ipmi = cij.env_to_dict(PREFIX, REQUIRED)

    if ipmi is None:
        ipmi["USER"] = "admin"
        ipmi["PASS"] = "admin"
        ipmi["HOST"] = "localhost"
        ipmi["PORT"] = "623"
        cij.info("ipmi.env: USER: %s, PASS: %s, HOST: %s, PORT: %s" % (
            ipmi["USER"], ipmi["PASS"], ipmi["HOST"], ipmi["PORT"]
        ))

    cij.env_export(PREFIX, EXPORTED, ipmi)

    return 0


def cmd(command):
    """Send IPMI 'command' via ipmitool"""

    env()

    ipmi = cij.env_to_dict(PREFIX, EXPORTED + REQUIRED)

    command = "ipmitool -U %s -P %s -H %s -p %s %s" % (
        ipmi["USER"], ipmi["PASS"], ipmi["HOST"], ipmi["PORT"], command)
    cij.info("ipmi.command: %s" % command)

    return cij.util.execute(command, shell=True, echo=True)


def pwr_on():
    """Target On"""

    cmd("power on")


def pwr_off():
    """Target Off"""

    cmd("power off")


def pwr_reset():
    """Target reset"""

    cmd("power reset")
